#!/usr/bin/env python
#
# Glances - An eye on your system
#
# SPDX-FileCopyrightText: 2025 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Glances API unitary tests suite."""

from unittest.mock import patch

from glances import __version__, api

# Global variables
# =================

# Init Glances API
# test_config = core.get_config()
# test_args = core.get_args()

testargs = ["glances", "-C", "./conf/glances.conf"]
with patch('sys.argv', testargs):
    gl = api.GlancesAPI()


# Pytest functions to test the Glances API version
def test_glances_api_version():
    assert gl.__version__ == __version__.split('.')[0]


def test_glances_api_plugins():
    # Check that the plugins list is not empty
    assert len(gl.plugins()) > 0
    # Check that the cpu plugin is in the list of plugins
    assert 'cpu' in gl.plugins()
    # Check that the network plugin is in the list of plugins
    assert 'network' in gl.plugins()
    # Check that the processcount plugin is in the list of plugins
    assert 'processcount' in gl.plugins()
    # Check that the processlist plugin is in the list of plugins
    assert 'processlist' in gl.plugins()


def test_glances_api_plugin_cpu():
    # Check that the cpu plugin is available
    assert gl.cpu is not None
    # Get list of keys (cpu stat fields)
    keys = gl.cpu.keys()
    # Check that the keys are not empty
    assert len(keys) > 0
    # Check that the cpu plugin has a total item
    assert gl.cpu['total'] is not None
    # and is a float
    assert isinstance(gl.cpu['total'], float)
    # test the get method
    assert isinstance(gl.cpu.get('total'), float)


def test_glances_api_plugin_network():
    # Check that the network plugin is available
    assert gl.network is not None
    # Get list of keys (interfaces)
    keys = gl.network.keys()
    # Check that the keys are not empty
    assert len(keys) > 0
    # Check that the first item is a dictionary
    assert isinstance(gl.network[keys[0]], dict)


def test_glances_api_plugin_process():
    # Get list of keys (processes)
    keys = gl.processcount.keys()
    # Check that the keys are not empty
    assert len(keys) > 0
    # Check that processcount total is > 0
    assert gl.processcount['total'] > 0

    # Get list of keys (processes)
    keys = gl.processlist.keys()
    # Check that first key is an integer (PID)
    assert isinstance(keys[0], int)
    # Check that the first item is a dictionary
    assert isinstance(gl.processlist[keys[0]], dict)


def test_glances_api_limits():
    assert isinstance(gl.cpu.limits, dict)
    assert isinstance(gl.network.limits, dict)



#
# This file is part of Glances.
#
# Copyright (C) 2018 Tim Nibert <docz2a@gmail.com>
# Copyright (C) 2026 Github@Drake7707
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""
Hard disk SMART attributes plugin.
Depends on pySMART and smartmontools
Must execute as root
"usermod -a -G disk USERNAME" is not sufficient unfortunately
SmartCTL (/usr/sbin/smartctl) must be in system path for python2.

Regular PySMART is a python2 library.
We are using the pySMART.smartx updated library to support both python 2 and 3.

If we only have disk group access (no root):
$ smartctl -i /dev/sda
smartctl 6.6 2016-05-31 r4324 [x86_64-linux-4.15.0-30-generic] (local build)
Copyright (C) 2002-16, Bruce Allen, Christian Franke, www.smartmontools.org


Probable ATA device behind a SAT layer
Try an additional '-d ata' or '-d sat' argument.

This is not very hopeful: https://medium.com/opsops/why-smartctl-could-not-be-run-without-root-7ea0583b1323

So, here is what we are going to do:
Check for admin access.  If no admin access, disable SMART plugin.

If smartmontools is not installed, we should catch the error upstream in plugin initialization.
"""

from glances.globals import is_admin
from glances.logger import logger
from glances.main import disable
from glances.plugins.plugin.model import GlancesPluginModel

# Import plugin specific dependency
try:
    from pySMART import DeviceList
    from pySMART.interface.nvme import NvmeAttributes
except ImportError as e:
    import_error_tag = True
    logger.warning(f"Missing Python Lib ({e}), HDD Smart plugin is disabled")
else:
    import_error_tag = False


def convert_attribute_to_dict(attr):
    return {
        'name': attr.name,
        'key': attr.name,
        'num': attr.num,
        'flags': attr.flags,
        'raw': attr.raw,
        'value': attr.value,
        'worst': attr.worst,
        'threshold': attr.thresh,
        'type': attr.type,
        'updated': attr.updated,
        'when_failed': attr.when_failed,
    }


# Keys for attributes that should be formatted with auto_unit (large byte values)
LARGE_VALUE_KEYS = frozenset(
    [
        "bytesWritten",
        "bytesRead",
        "dataUnitsRead",
        "dataUnitsWritten",
        "hostReadCommands",
        "hostWriteCommands",
    ]
)

NVME_ATTRIBUTE_LABELS = {
    "criticalWarning": "Number of critical warnings",
    "_temperature": "Temperature (°C)",
    "availableSpare": "Available spare (%)",
    "availableSpareThreshold": "Available spare threshold (%)",
    "percentageUsed": "Percentage used (%)",
    "dataUnitsRead": "Data units read",
    "bytesRead": "Bytes read",
    "dataUnitsWritten": "Data units written",
    "bytesWritten": "Bytes written",
    "hostReadCommands": "Host read commands",
    "hostWriteCommands": "Host write commands",
    "controllerBusyTime": "Controller busy time (min)",
    "powerCycles": "Power cycles",
    "powerOnHours": "Power-on hours",
    "unsafeShutdowns": "Unsafe shutdowns",
    "integrityErrors": "Integrity errors",
    "errorEntries": "Error log entries",
    "warningTemperatureTime": "Warning temperature time (min)",
    "criticalTemperatureTime": "Critical temperature time (min)",
    "_logical_sector_size": "Logical sector size",
    "_physical_sector_size": "Physical sector size",
    "errors": "Errors",
    "tests": "Self-tests",
}


def convert_nvme_attribute_to_dict(key, value):
    label = NVME_ATTRIBUTE_LABELS.get(key, key)

    return {
        'name': label,
        'key': key,
        'value': value,
        'flags': None,
        'raw': value,
        'worst': None,
        'threshold': None,
        'type': None,
        'updated': None,
        'when_failed': None,
    }


def _process_standard_attributes(device_stats, attributes, hide_attributes):
    """Process standard SMART attributes and add them to device_stats."""
    for attribute in attributes:
        if attribute is None or attribute.name in hide_attributes:
            continue

        attrib_dict = convert_attribute_to_dict(attribute)
        num = attrib_dict.pop('num', None)
        if num is None:
            logger.debug(f'Smart plugin error - Skip attribute with no num: {attribute}')
            continue

        device_stats[num] = attrib_dict


def _process_nvme_attributes(device_stats, if_attributes, hide_attributes):
    """Process NVMe-specific attributes and add them to device_stats."""
    if not isinstance(if_attributes, NvmeAttributes):
        return

    for idx, (attr, value) in enumerate(vars(if_attributes).items(), start=1):
        attrib_dict = convert_nvme_attribute_to_dict(attr, value)
        if attrib_dict['name'] in hide_attributes:
            continue

        # Verify the value is serializable to prevent rendering errors
        if value is not None:
            try:
                str(value)
            except Exception:
                logger.debug(f'Unable to serialize attribute {attr} from NVME')
                attrib_dict['value'] = None
                attrib_dict['raw'] = None

        device_stats[idx] = attrib_dict


def get_smart_data(hide_attributes):
    """Get SMART attribute data.

    Returns a list of dictionaries, each containing:
    - 'DeviceName': Device identification string
    - Numeric keys: SMART attribute dictionaries with flags, raw values, etc.
    """
    stats = []
    try:
        devlist = DeviceList()
    except TypeError as e:
        logger.debug(f'Smart plugin error - Can not grab device list ({e})')
        global import_error_tag
        import_error_tag = True
        return stats

    for dev in devlist.devices:
        device_stats = {'DeviceName': f'{dev.name} {dev.model}'}
        _process_standard_attributes(device_stats, dev.attributes, hide_attributes)
        _process_nvme_attributes(device_stats, dev.if_attributes, hide_attributes)
        stats.append(device_stats)

    return stats


class SmartPlugin(GlancesPluginModel):
    """Glances' HDD SMART plugin."""

    def __init__(self, args=None, config=None, stats_init_value=[]):
        """Init the plugin."""
        if not is_admin() and args:
            disable(args, "smart")
            logger.debug("Current user is not admin, HDD SMART plugin disabled.")

        super().__init__(args=args, config=config)

        self.display_curse = True
        self.hide_attributes = self._parse_hide_attributes(config)

    def _parse_hide_attributes(self, config):
        """Parse and return the list of attributes to hide from config."""
        smart_config = config.as_dict().get('smart', {})
        hide_attr_str = smart_config.get('hide_attributes', '')
        if hide_attr_str:
            logger.info(f'Following SMART attributes will not be displayed: {hide_attr_str}')
            return hide_attr_str.split(',')
        return []

    @property
    def hide_attributes(self):
        """Set hide_attributes list"""
        return self._hide_attributes

    @hide_attributes.setter
    def hide_attributes(self, attr_list):
        """Set hide_attributes list"""
        self._hide_attributes = list(attr_list)

    @GlancesPluginModel._check_decorator
    @GlancesPluginModel._log_result_decorator
    def update(self):
        """Update SMART stats using the input method."""
        # Init new stats
        stats = self.get_init_value()

        if import_error_tag:
            return self.stats

        if self.input_method == 'local':
            # Update stats and hide some sensors(#2996)
            stats = [
                dict(s, key=self.get_key())
                for s in get_smart_data(self.hide_attributes)
                if self.is_display(s[self.get_key()])
            ]
        elif self.input_method == 'snmp':
            pass

        # Update the stats
        self.stats = stats

        return self.stats

    def get_key(self):
        """Return the key of the list."""
        return 'DeviceName'

    def _format_raw_value(self, stat):
        """Format a raw SMART attribute value for display."""
        raw = stat['raw']
        if raw is None:
            return ""
        if stat['key'] in LARGE_VALUE_KEYS:
            return self.auto_unit(raw)
        return str(raw)

    def _get_sorted_stat_keys(self, device_stat):
        """Get sorted attribute keys from device stats, excluding DeviceName."""
        keys = [k for k in device_stat if k != 'DeviceName']
        try:
            return sorted(keys, key=int)
        except ValueError:
            # Some keys may not be numeric (see #2904)
            return keys

    def _add_device_stats(self, ret, device_stat, max_width, name_max_width):
        """Add a device's SMART stats to the curse output."""
        ret.append(self.curse_new_line())
        ret.append(self.curse_add_line(f'{device_stat["DeviceName"][:max_width]:{max_width}}'))

        for key in self._get_sorted_stat_keys(device_stat):
            stat = device_stat[key]
            ret.append(self.curse_new_line())

            # Attribute name
            name = stat['name'][: name_max_width - 1].replace('_', ' ')
            ret.append(self.curse_add_line(f' {name:{name_max_width - 1}}'))

            # Attribute value
            try:
                value_str = self._format_raw_value(stat)
                ret.append(self.curse_add_line(f'{value_str:>8}'))
            except Exception:
                logger.debug(f"Failed to serialize {key}")
                ret.append(self.curse_add_line(""))

    def msg_curse(self, args=None, max_width=None):
        """Return the dict to display in the curse interface."""
        ret = []

        if import_error_tag or not self.stats or self.is_disabled():
            return ret

        if not max_width:
            logger.debug(f"No max_width defined for the {self.plugin_name} plugin, it will not be displayed.")
            return ret

        name_max_width = max_width - 6

        # Header
        ret.append(self.curse_add_line(f'{"SMART disks":{name_max_width}}', "TITLE"))

        # Device data
        for device_stat in self.stats:
            self._add_device_stats(ret, device_stat, max_width, name_max_width)

        return ret



#!/usr/bin/env python
#
# Glances - An eye on your system
#
# SPDX-FileCopyrightText: 2026 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Tests for the Ports plugin."""

import pytest

from glances.plugins.ports import PortsPlugin


@pytest.fixture
def ports_plugin():
    """Return a Ports plugin instance without running its full init."""
    return PortsPlugin.__new__(PortsPlugin)


def web_scan(status, elapsed=0, rtt_warning=1):
    """Return a web scan result as stored by the ports plugin."""
    return {'status': status, 'elapsed': elapsed, 'rtt_warning': rtt_warning}


class TestPortsPluginAlertLevel:
    """Test that the alert level is resolved by severity, not by dict ordering."""

    @pytest.mark.parametrize(
        ('conds', 'expected'),
        [
            ({'CAREFUL': True, 'CRITICAL': True, 'WARNING': True}, 'CRITICAL'),
            ({'CAREFUL': True, 'CRITICAL': False, 'WARNING': True}, 'WARNING'),
            ({'CAREFUL': True, 'CRITICAL': False, 'WARNING': False}, 'CAREFUL'),
            ({'CAREFUL': False, 'CRITICAL': False, 'WARNING': False}, 'OK'),
        ],
    )
    def test_most_severe_condition_wins(self, ports_plugin, conds, expected):
        """Test that the most severe matching condition is returned."""
        assert ports_plugin.get_default_ret_value(conds) == expected

    def test_web_not_scanned_yet_is_careful(self, ports_plugin):
        """Test that a URL whose first scan did not complete is not CRITICAL."""
        conds = ports_plugin.get_conds_if_url(web_scan(None))
        assert ports_plugin.get_default_ret_value(conds) == 'CAREFUL'

    def test_web_failing_and_slow_is_critical(self, ports_plugin):
        """Test that a failing URL stays CRITICAL even when it is also slow."""
        conds = ports_plugin.get_conds_if_url(web_scan(404, elapsed=5))
        assert ports_plugin.get_default_ret_value(conds) == 'CRITICAL'

    def test_web_failing_and_fast_is_critical(self, ports_plugin):
        """Test that a failing URL is CRITICAL."""
        conds = ports_plugin.get_conds_if_url(web_scan(404))
        assert ports_plugin.get_default_ret_value(conds) == 'CRITICAL'

    def test_web_ok_but_slow_is_warning(self, ports_plugin):
        """Test that a reachable but slow URL is WARNING."""
        conds = ports_plugin.get_conds_if_url(web_scan(200, elapsed=5))
        assert ports_plugin.get_default_ret_value(conds) == 'WARNING'

    def test_web_ok_and_fast_is_ok(self, ports_plugin):
        """Test that a reachable and fast URL is OK."""
        conds = ports_plugin.get_conds_if_url(web_scan(200))
        assert ports_plugin.get_default_ret_value(conds) == 'OK'



import os
import sys
import time

import psutil

sys.path.insert(0, '../glances')

from glances.globals import disable, enable
from glances.main import GlancesMain
from glances.stats import GlancesStats

core = GlancesMain()
stats = GlancesStats(config=core.get_config(), args=core.get_args())

refresh = 2
iteration = 6
pid = os.getpid()
process = psutil.Process(pid)

print("Check Glances sensors plugin CPU consumption overhead")
print("=====================================================")
print()

print(f"Init Glances to make the stats more relevant (wait {iteration * refresh} seconds)")
for i in range(0, iteration):
    stats.update()
    time.sleep(refresh)

print("CPU consumption with sensors plugin disable")
disable(stats.get_plugin('sensors').args, 'sensors')
stats.get_plugin('sensors').reset()
cpu_start = sum(process.cpu_times()[:2])
for i in range(0, iteration):
    stats.update()
    print(f'{i + 1}/{iteration}: {len(stats.get_plugin("sensors").get_raw())} sensor')
    time.sleep(refresh)
cpu_end = sum(process.cpu_times()[:2])
cpu_without_sensors = cpu_end - cpu_start
print(f'Glances process consumption (user + kernel) with sensors plugin disable: {cpu_without_sensors}')

print("CPU consumption with sensors plugin enable")
enable(stats.get_plugin('sensors').args, 'sensors')
stats.get_plugin('sensors').reset()
cpu_start = sum(process.cpu_times()[:2])
for i in range(0, iteration):
    stats.update()
    print(f'{i + 1}/{iteration}: {len(stats.get_plugin("sensors").get_raw())} sensors')
    time.sleep(refresh)
cpu_end = sum(process.cpu_times()[:2])
cpu_with_sensors = cpu_end - cpu_start
print(f'Glances process consumption (user + kernel) with sensors plugin enable: {cpu_with_sensors}')

print(
    f'Percentage of CPU consumption increase with sensors plugin enable: \
 {((cpu_with_sensors - cpu_without_sensors) / cpu_without_sensors) * 100:.2f}%'
)



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2022 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Riemann interface class."""

import socket
from numbers import Number

# Import bernhard for Riemann
import bernhard

from glances.exports.export import GlancesExport
from glances.logger import logger


class Export(GlancesExport):
    """This class manages the Riemann export module."""

    def __init__(self, config=None, args=None):
        """Init the Riemann export IF."""
        super().__init__(config=config, args=args)

        # Mandatory configuration keys (additional to host and port)
        # N/A

        # Optional configuration keys
        # N/A

        # Load the Riemann configuration
        self.export_enable = self.load_conf('riemann', mandatories=['host', 'port'], options=[])
        if not self.export_enable:
            exit('Missing RIEMANN config')

        # Get the current hostname
        self.hostname = socket.gethostname()

        # Init the Riemann client
        self.client = self.init()

    def init(self):
        """Init the connection to the Riemann server."""
        if not self.export_enable:
            return None
        try:
            return bernhard.Client(host=self.host, port=self.port)
        except Exception as e:
            logger.critical(f"Connection to Riemann failed : {e} ")
            return None

    def export(self, name, columns, points):
        """Write the points in Riemann."""
        for i in range(len(columns)):
            if not isinstance(points[i], Number):
                continue

            data = {'host': self.hostname, 'service': name + " " + columns[i], 'metric': points[i]}
            logger.debug(data)
            try:
                self.client.send(data)
            except Exception as e:
                logger.error(f"Cannot export stats to Riemann ({e})")



#
# This file is part of Glances.
#
# Copyright (C) 2020 Kirby Banman <kirby.banman@gmail.com>
# Copyright (C) 2024 Nicolas Hennion <nicolashennion@gmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""GPU plugin for Glances.

Currently supported:
- NVIDIA GPU (need pynvml lib)
- AMD GPU (no lib needed)
- Intel GPU (no lib needed)
- ARM GPU (Linux kernel >= 6.0, no lib needed; drivers: msm/Adreno,
  Panfrost, Panthor, v3d, Lima, Etnaviv)

Quick test:
- Start Glances
- In a terminal: vblank_mode=0 glxgears
"""

import glances.gpu_percent as gpu_avg_stats
from glances.globals import to_fahrenheit
from glances.logger import logger
from glances.plugins.gpu.cards.amd import AmdGPU
from glances.plugins.gpu.cards.arm import ArmGPU
from glances.plugins.gpu.cards.intel import IntelGPU
from glances.plugins.gpu.cards.nvidia import NvidiaGPU
from glances.plugins.plugin.model import GlancesPluginModel

# Fields description
# description: human readable description
# short_name: shortname to use un UI
# unit: unit type
# rate: is it a rate ? If yes, // by time_since_update when displayed,
# min_symbol: Auto unit should be used if value > than 1 'X' (K, M, G)...
fields_description = {
    'gpu_id': {
        'description': 'GPU identification',
    },
    'name': {
        'description': 'GPU name',
    },
    'mem': {
        'description': 'Memory consumption',
        'unit': 'percent',
    },
    'proc': {
        'description': 'GPU processor consumption',
        'unit': 'percent',
    },
    'temperature': {
        'description': 'GPU temperature',
        'unit': 'celsius',
    },
    'fan_speed': {
        'description': 'GPU fan speed',
        'unit': 'roundperminute',
    },
}

# Define the history items list
# All items in this list will be historised if the --enable-history tag is set
items_history_list = [
    {'name': 'proc', 'description': 'GPU processor', 'y_unit': '%'},
    {'name': 'mem', 'description': 'Memory consumption', 'y_unit': '%'},
]


class GpuPlugin(GlancesPluginModel):
    """Glances GPU plugin.

    stats is a list of dictionaries with one entry per GPU
    """

    def __init__(self, args=None, config=None):
        """Init the plugin."""
        super().__init__(
            args=args,
            config=config,
            items_history_list=items_history_list,
            stats_init_value=[],
            fields_description=fields_description,
        )
        # Init the Nvidia GPU API
        try:
            self.nvidia = NvidiaGPU()
        except Exception as e:
            logger.debug(f'Nvidia GPU initialization error: {e}')
            self.nvidia = None

        # Init the AMD GPU API
        try:
            self.amd = AmdGPU()
        except Exception as e:
            logger.debug(f'AMD GPU initialization error: {e}')
            self.amd = None
        # Just for test purpose (uncomment to test on computer without AMD GPU)
        # self.amd = AmdGPU(drm_root_folder='./tests-data/plugins/gpu/amd/sys/class/drm')

        # Init the Intel GPU API
        try:
            self.intel = IntelGPU()
        except Exception as e:
            logger.debug(f'Intel GPU initialization error: {e}')
            self.intel = None
        # Just for test purpose (uncomment to test on computer without Intel GPU)
        # self.intel = IntelGPU(drm_root_folder='./tests-data/plugins/gpu/intel/sys/class/drm')

        # Init the ARM GPU API (msm, panfrost, panthor, v3d, lima, etnaviv)
        try:
            self.arm = ArmGPU()
        except Exception as e:
            logger.debug(f'ARM GPU initialization error: {e}')
            self.arm = None
        # Just for test purpose (uncomment to test on computer without ARM GPU)
        # self.arm = ArmGPU(
        #     drm_root_folder='./tests-data/plugins/gpu/arm/sys/class/drm',
        #     proc_root_folder='./tests-data/plugins/gpu/arm/proc',
        # )

        # We want to display the stat in the curse interface
        self.display_curse = True

    def exit(self):
        """Overwrite the exit method to close the GPU API."""
        # Each backend may be None if its initialization failed.
        if self.nvidia:
            self.nvidia.exit()
        if self.amd:
            self.amd.exit()
        if self.intel:
            self.intel.exit()
        if self.arm:
            self.arm.exit()
        # Call the father exit method
        super().exit()

    def get_key(self):
        """Return the key of the list."""
        return 'gpu_id'

    @GlancesPluginModel._check_decorator
    @GlancesPluginModel._log_result_decorator
    def update(self):
        """Update the GPU stats."""
        # Init new stats
        stats = self.get_init_value()

        # Get the stats
        if self.nvidia:
            stats.extend(self.nvidia.get_device_stats())
        if self.amd:
            stats.extend(self.amd.get_device_stats())
        if self.intel:
            stats.extend(self.intel.get_device_stats())
        if self.arm:
            stats.extend(self.arm.get_device_stats())

        # !!!
        # Uncomment to test on computer without Nvidia GPU
        # One GPU sample:
        # stats = [
        #     {
        #         "key": "gpu_id",
        #         "gpu_id": "nvidia0",
        #         "name": "Fake GeForce GTX",
        #         "mem": 5.792331695556641,
        #         "proc": 4,
        #         "temperature": 26,
        #         "fan_speed": 30,
        #     }
        # ]
        # Two GPU sample:
        # stats = [
        #     {
        #         "key": "gpu_id",
        #         "gpu_id": "nvidia0",
        #         "name": "Fake GeForce GTX1",
        #         "mem": 5.792331695556641,
        #         "proc": 4,
        #         "temperature": 26,
        #         "fan_speed": 30,
        #     },
        #     {
        #         "key": "gpu_id",
        #         "gpu_id": "nvidia1",
        #         "name": "Fake GeForce GTX1",
        #         "mem": 15,
        #         "proc": 8,
        #         "temperature": 65,
        #         "fan_speed": 75,
        #     },
        # ]
        if gpu_avg_stats.get_gpu_mem:
            avg_mem = self._get_mean("mem")
            if avg_mem is None:
                avg_mem = 0
            gpu_avg_stats.gpu_mem = avg_mem

        if gpu_avg_stats.get_gpu_proc:
            avg_proc = self._get_mean("proc")
            if avg_proc is None:
                avg_proc = 0
            gpu_avg_stats.gpu_proc = avg_proc

        # Update the stats
        self.stats = stats

        return self.stats

    def update_views(self):
        """Update stats views."""
        # Call the father's method
        super().update_views()

        # Add specifics information
        # Alert
        for i in self.stats:
            # Init the views for the current GPU
            self.views[i[self.get_key()]] = {'proc': {}, 'mem': {}, 'temperature': {}}
            # Processor alert
            if 'proc' in i:
                alert = self.get_alert(i['proc'], header='proc')
                self.views[i[self.get_key()]]['proc']['decoration'] = alert
            # Memory alert
            if 'mem' in i:
                alert = self.get_alert(i['mem'], header='mem')
                self.views[i[self.get_key()]]['mem']['decoration'] = alert
            # Temperature alert
            if 'temperature' in i:
                alert = self.get_alert(i['temperature'], header='temperature')
                self.views[i[self.get_key()]]['temperature']['decoration'] = alert

        return True

    def _get_mean(self, key):
        """Calculate mean value for a given key across all GPU stats.

        Returns None if calculation fails (e.g., missing data).
        """
        try:
            return sum(s[key] for s in self.stats if s is not None) / len(self.stats)
        except (TypeError, ZeroDivisionError):
            return None

    def _format_value(self, value, unit='%'):
        """Format a value with unit, or return N/A if value is None."""
        if value is None:
            return '{:>4}'.format('N/A')
        return f'{value:>3.0f}{unit}'

    def _build_header(self):
        """Build the header string based on GPU count and names."""
        same_name = all(s['name'] == self.stats[0]['name'] for s in self.stats)
        gpu_name = self.stats[0]['name']
        gpu_count = len(self.stats)

        if gpu_count > 1:
            if same_name:
                header = f'{gpu_count} {gpu_name}'
            else:
                header = f'{gpu_count} GPUs'
        elif same_name:
            header = gpu_name
        else:
            header = 'GPU'
        return header[:17]

    def _add_metric_line(self, ret, key, label, label_mean, unit='%'):
        """Add a metric line (label + value) to the curse output.

        Args:
            ret: The return list to append to
            key: The metric key (e.g., 'proc', 'mem', 'temperature')
            label: Label for single GPU mode
            label_mean: Label for multi-GPU mean mode
            unit: Unit suffix for the value (default: '%')
        """
        gpu_stats = self.stats[0]
        is_multi = len(self.stats) > 1

        ret.append(self.curse_new_line())
        ret.append(self.curse_add_line(f'{label_mean if is_multi else label:13}'))
        mean_value = self._get_mean(key)
        ret.append(
            self.curse_add_line(
                self._format_value(mean_value, unit),
                self.get_views(item=gpu_stats[self.get_key()], key=key, option='decoration'),
            )
        )

    def _msg_curse_summary(self, ret, args):
        """Build curse output for summary view (single GPU or mean mode)."""
        self._add_metric_line(ret, 'proc', 'proc:', 'proc mean:')
        self._add_metric_line(ret, 'mem', 'mem:', 'mem mean:')

        # Temperature needs special handling for Fahrenheit conversion
        gpu_stats = self.stats[0]
        is_multi = len(self.stats) > 1
        ret.append(self.curse_new_line())
        ret.append(self.curse_add_line('{:13}'.format('temp mean:' if is_multi else 'temperature:')))
        mean_temp = self._get_mean('temperature')
        if mean_temp is not None and args and args.fahrenheit:
            mean_temp = to_fahrenheit(mean_temp)
        unit = 'F' if args and args.fahrenheit else 'C'
        ret.append(
            self.curse_add_line(
                self._format_value(mean_temp, unit),
                self.get_views(item=gpu_stats[self.get_key()], key='temperature', option='decoration'),
            )
        )

    def _msg_curse_multi(self, ret):
        """Build curse output for multi-GPU detailed view."""
        for gpu_stats in self.stats:
            ret.append(self.curse_new_line())
            # id_msg = '{:7}'.format(gpu_stats['gpu_id'])
            id_msg = '{:7}'.format(gpu_stats['name'][0:9])
            ret.append(self.curse_add_line(f'{id_msg}'))
            # Each metric is a standalone segment: it carries its own decoration and must not
            # re-emit the previous one. An unavailable sensor is displayed as N/A, not hidden.
            ret.append(
                self.curse_add_line(
                    f" {self._format_value(gpu_stats.get('proc'))}",
                    self.get_views(item=gpu_stats[self.get_key()], key='proc', option='decoration'),
                )
            )
            ret.append(
                self.curse_add_line(
                    f" mem {self._format_value(gpu_stats.get('mem'))}",
                    self.get_views(item=gpu_stats[self.get_key()], key='mem', option='decoration'),
                )
            )

    def msg_curse(self, args=None, max_width=None):
        """Return the dict to display in the curse interface."""
        ret = []

        # Only process if stats exist, not empty (issue #871) and plugin not disabled
        if not self.stats or self.is_disabled():
            return ret

        # Header
        ret.append(self.curse_add_line(self._build_header(), "TITLE"))

        # Build the string message
        if len(self.stats) == 1 or args.meangpu:
            self._msg_curse_summary(ret, args)
        else:
            self._msg_curse_multi(ret)

        return ret



#!/usr/bin/env python
#
# Glances - An eye on your system
#
# SPDX-FileCopyrightText: 2024 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Glances unit tests for the TUI Central Browser mode.

Tests cover:
- Config generation from the default glances.conf with browser options enabled
- Static server list loading from config
- Password list loading and lookup (host-specific, default, missing)
- URI generation (with/without credentials, static vs dynamic servers)
- Credential sanitization: dynamic (Zeroconf) servers must not inherit saved passwords
- Credential sanitization: URIs for dynamic servers use IP, not advertised name
"""

import re
from pathlib import Path
from unittest.mock import patch

import pytest

from glances.config import Config
from glances.password_list import GlancesPasswordList
from glances.servers_list import GlancesServersList

# ---------------------------------------------------------------------------
# Helpers – generate a browser-enabled config from the default glances.conf
# ---------------------------------------------------------------------------

DEFAULT_CONF = Path(__file__).resolve().parent.parent / 'conf' / 'glances.conf'

# Servers injected into the generated config
# All names must be DNS-resolvable (localhost, 127.0.0.1) to avoid gethostbyname errors
STATIC_SERVERS = [
    {'name': 'localhost', 'alias': 'Local RPC Server', 'port': '61209', 'protocol': 'rpc'},
    {'name': 'localhost', 'alias': 'Local REST Server', 'port': '61208', 'protocol': 'rest'},
    {'name': '127.0.0.1', 'alias': 'Loopback Server', 'port': '61210', 'protocol': 'rpc'},
]

PASSWORDS = {
    'localhost': 'localpwd',
    '127.0.0.1': 'loopbackpwd',
    'default': 'defaultpassword',
}


def _generate_browser_conf(tmp_path):
    """Read the default glances.conf, uncomment and populate [serverlist] and
    [passwords] sections, write the result to *tmp_path* and return its path.

    This is regenerated at every test run so the test stays in sync with the
    upstream default config without manual maintenance.
    """
    source = DEFAULT_CONF.read_text(encoding='utf-8')

    # --- Build [serverlist] replacement ---
    serverlist_lines = [
        '[serverlist]',
        'columns=system:hr_name,load:min5,cpu:total,mem:percent',
    ]
    for idx, srv in enumerate(STATIC_SERVERS, start=1):
        serverlist_lines.append(f'server_{idx}_name={srv["name"]}')
        serverlist_lines.append(f'server_{idx}_alias={srv["alias"]}')
        serverlist_lines.append(f'server_{idx}_port={srv["port"]}')
        serverlist_lines.append(f'server_{idx}_protocol={srv["protocol"]}')
    serverlist_block = '\n'.join(serverlist_lines) + '\n'

    # --- Build [passwords] replacement ---
    password_lines = ['[passwords]']
    for host, pwd in PASSWORDS.items():
        password_lines.append(f'{host}={pwd}')
    password_block = '\n'.join(password_lines) + '\n'

    # Replace existing sections in-place
    # Match from section header to right before the next section header (or EOF)
    source = re.sub(
        r'\[serverlist\].*?(?=\n\[|\Z)',
        serverlist_block,
        source,
        count=1,
        flags=re.DOTALL,
    )
    source = re.sub(
        r'\[passwords\].*?(?=\n\[|\Z)',
        password_block,
        source,
        count=1,
        flags=re.DOTALL,
    )

    conf_file = tmp_path / 'glances.conf'
    conf_file.write_text(source, encoding='utf-8')
    return str(conf_file)


# ---------------------------------------------------------------------------
# Fixtures
# ---------------------------------------------------------------------------


@pytest.fixture(scope='module')
def browser_conf_path(tmp_path_factory):
    """Generate a browser-enabled config file once per module."""
    return _generate_browser_conf(tmp_path_factory.mktemp('browser_conf'))


@pytest.fixture(scope='module')
def browser_config(browser_conf_path):
    """Return a Config object loaded from the generated browser config."""
    return Config(browser_conf_path)


@pytest.fixture(scope='module')
def browser_args(browser_conf_path):
    """Return a minimal args namespace suitable for GlancesServersList."""
    from glances.main import GlancesMain

    testargs = ['glances', '--browser', '--disable-autodiscover', '-C', browser_conf_path]
    with patch('sys.argv', testargs):
        core = GlancesMain()
    return core.get_args()


@pytest.fixture(scope='module')
def servers_list(browser_config, browser_args):
    """Return a fully initialised GlancesServersList (autodiscovery disabled)."""
    return GlancesServersList(config=browser_config, args=browser_args)


# ---------------------------------------------------------------------------
# Tests – Config generation
# ---------------------------------------------------------------------------


class TestBrowserConfigGeneration:
    """Verify the generated config has the expected sections and values."""

    def test_serverlist_section_exists(self, browser_config):
        assert browser_config.has_section('serverlist')

    def test_passwords_section_exists(self, browser_config):
        assert browser_config.has_section('passwords')

    def test_server_count(self, browser_config):
        """All defined servers should be loadable."""
        count = 0
        for i in range(1, 256):
            if browser_config.get_value('serverlist', f'server_{i}_name') is not None:
                count += 1
        assert count == len(STATIC_SERVERS)

    def test_password_values(self, browser_config):
        items = dict(browser_config.items('passwords'))
        for host, pwd in PASSWORDS.items():
            assert items[host] == pwd


# ---------------------------------------------------------------------------
# Tests – Static server list loading
# ---------------------------------------------------------------------------


class TestStaticServerList:
    """Verify GlancesStaticServer correctly loads servers from config."""

    def test_server_list_length(self, servers_list):
        servers = servers_list.get_servers_list()
        assert len(servers) == len(STATIC_SERVERS)

    def test_server_fields(self, servers_list):
        """Each server dict must have the expected keys."""
        required_keys = {'key', 'name', 'ip', 'port', 'protocol', 'username', 'password', 'status', 'type'}
        for server in servers_list.get_servers_list():
            assert required_keys.issubset(server.keys()), f"Missing keys in {server}"

    def test_server_names(self, servers_list):
        names = [s['name'] for s in servers_list.get_servers_list()]
        for srv in STATIC_SERVERS:
            assert srv['name'] in names

    def test_server_protocols(self, servers_list):
        for server in servers_list.get_servers_list():
            assert server['protocol'] in ('rpc', 'rest')

    def test_server_type_is_static(self, servers_list):
        for server in servers_list.get_servers_list():
            assert server['type'] == 'STATIC'

    def test_server_initial_status(self, servers_list):
        for server in servers_list.get_servers_list():
            assert server['status'] == 'UNKNOWN'

    def test_server_default_username(self, servers_list):
        for server in servers_list.get_servers_list():
            assert server['username'] == 'glances'

    def test_server_default_empty_password(self, servers_list):
        for server in servers_list.get_servers_list():
            assert server['password'] == ''


# ---------------------------------------------------------------------------
# Tests – Password list
# ---------------------------------------------------------------------------


class TestPasswordList:
    """Verify GlancesPasswordList loading and lookup."""

    def test_host_specific_password(self, servers_list):
        assert servers_list.password.get_password('localhost') == 'localpwd'

    def test_loopback_password(self, servers_list):
        assert servers_list.password.get_password('127.0.0.1') == 'loopbackpwd'

    def test_default_fallback(self, servers_list):
        """Unknown host should fall back to 'default' password."""
        assert servers_list.password.get_password('unknown-host') == 'defaultpassword'

    def test_no_password_without_default(self):
        """Without a default entry, unknown host should return None."""
        pwd = GlancesPasswordList()
        pwd._password_dict = {'localhost': 'localpwd'}
        assert pwd.get_password('unknown-host') is None


# ---------------------------------------------------------------------------
# Tests – Columns
# ---------------------------------------------------------------------------


class TestColumnsDefinition:
    """Verify columns parsing from config."""

    def test_columns_loaded(self, servers_list):
        columns = servers_list.get_columns()
        assert len(columns) > 0

    def test_columns_structure(self, servers_list):
        for col in servers_list.get_columns():
            assert 'plugin' in col
            assert 'field' in col

    def test_columns_values(self, servers_list):
        plugins = [c['plugin'] for c in servers_list.get_columns()]
        assert 'system' in plugins
        assert 'cpu' in plugins
        assert 'mem' in plugins


# ---------------------------------------------------------------------------
# Tests – URI generation for STATIC servers
# ---------------------------------------------------------------------------


class TestGetUriStatic:
    """Verify URI generation for static servers."""

    def test_uri_without_password(self, servers_list):
        """Static server with empty password → URI without credentials."""
        server = {
            'name': 'myhost',
            'ip': '10.0.0.1',
            'port': 61209,
            'username': 'glances',
            'password': '',
            'status': 'ONLINE',
            'type': 'STATIC',
        }
        uri = servers_list.get_uri(server)
        assert uri == 'http://myhost:61209'
        assert '@' not in uri

    def test_uri_with_password(self, servers_list):
        """Static server with a non-empty password → URI with credentials."""
        server = {
            'name': 'myhost',
            'ip': '10.0.0.1',
            'port': 61209,
            'username': 'glances',
            'password': 'somehash',
            'status': 'ONLINE',
            'type': 'STATIC',
        }
        uri = servers_list.get_uri(server)
        assert 'glances:somehash@myhost:61209' in uri

    def test_uri_protected_uses_saved_password(self, servers_list):
        """PROTECTED static server with a saved password should get the hash injected."""
        server = {
            'name': 'localhost',
            'ip': '127.0.0.1',
            'port': 61209,
            'username': 'glances',
            'password': 'placeholder',
            'status': 'PROTECTED',
            'type': 'STATIC',
        }
        uri = servers_list.get_uri(server)
        # Password should have been replaced by the hash of 'localpwd'
        expected_hash = servers_list.password.get_hash('localpwd')
        assert expected_hash in uri
        assert server['password'] == expected_hash

    def test_uri_protected_default_fallback(self, servers_list):
        """PROTECTED static server with unknown name falls back to 'default' password."""
        server = {
            'name': 'unknown-static-host',
            'ip': '10.0.0.2',
            'port': 61209,
            'username': 'glances',
            'password': 'placeholder',
            'status': 'PROTECTED',
            'type': 'STATIC',
        }
        uri = servers_list.get_uri(server)
        expected_hash = servers_list.password.get_hash('defaultpassword')
        assert expected_hash in uri

    def test_uri_static_uses_name_not_ip(self, servers_list):
        """Static server URI should use server['name'] as the host."""
        server = {
            'name': 'myhost.local',
            'ip': '192.168.1.100',
            'port': 61209,
            'username': 'glances',
            'password': '',
            'status': 'ONLINE',
            'type': 'STATIC',
        }
        uri = servers_list.get_uri(server)
        assert 'myhost.local' in uri
        assert '192.168.1.100' not in uri


# ---------------------------------------------------------------------------
# Tests – URI generation for DYNAMIC servers (CVE fix validation)
# ---------------------------------------------------------------------------


class TestGetUriDynamic:
    """Verify that dynamic (Zeroconf) servers use IP and never inherit saved passwords.

    These tests validate the fix for the Zeroconf credential exfiltration CVE.
    """

    def test_uri_dynamic_uses_ip_not_name(self, servers_list):
        """DYNAMIC server URI must use the discovered IP, not the advertised name."""
        server = {
            'name': 'attacker-controlled-name',
            'ip': '192.168.1.50',
            'port': 61209,
            'username': 'glances',
            'password': '',
            'status': 'ONLINE',
            'type': 'DYNAMIC',
        }
        uri = servers_list.get_uri(server)
        assert '192.168.1.50' in uri
        assert 'attacker-controlled-name' not in uri

    def test_uri_dynamic_protected_no_saved_password(self, servers_list):
        """DYNAMIC PROTECTED server must NOT inherit saved or default passwords."""
        server = {
            'name': 'localhost',  # name matches a saved password
            'ip': '198.51.100.50',
            'port': 61209,
            'username': 'glances',
            'password': 'placeholder',
            'status': 'PROTECTED',
            'type': 'DYNAMIC',
        }
        uri = servers_list.get_uri(server)
        # The password should remain as 'placeholder' (not replaced by saved hash)
        assert server['password'] == 'placeholder'
        # URI should use IP
        assert '198.51.100.50' in uri
        assert 'localhost' not in uri

    def test_uri_dynamic_no_default_password_fallback(self, servers_list):
        """DYNAMIC server with unknown name must NOT fall back to default password."""
        server = {
            'name': 'fake-zeroconf-service',
            'ip': '198.51.100.99',
            'port': 61209,
            'username': 'glances',
            'password': 'placeholder',
            'status': 'PROTECTED',
            'type': 'DYNAMIC',
        }
        uri = servers_list.get_uri(server)
        # password unchanged
        assert server['password'] == 'placeholder'
        # No hash of 'defaultpassword' should appear
        default_hash = servers_list.password.get_hash('defaultpassword')
        assert default_hash not in uri

    def test_get_connect_host_static(self, servers_list):
        server = {'name': 'myhost', 'ip': '10.0.0.1', 'type': 'STATIC'}
        assert servers_list._get_connect_host(server) == 'myhost'

    def test_get_connect_host_dynamic(self, servers_list):
        server = {'name': 'advertised-name', 'ip': '10.0.0.2', 'type': 'DYNAMIC'}
        assert servers_list._get_connect_host(server) == '10.0.0.2'

    def test_get_preconfigured_password_static(self, servers_list):
        server = {'name': 'localhost', 'type': 'STATIC'}
        assert servers_list._get_preconfigured_password(server) == 'localpwd'

    def test_get_preconfigured_password_dynamic_returns_none(self, servers_list):
        server = {'name': 'localhost', 'type': 'DYNAMIC'}
        assert servers_list._get_preconfigured_password(server) is None

    def test_get_preconfigured_password_dynamic_no_default(self, servers_list):
        """Even with a 'default' password configured, dynamic servers get None."""
        server = {'name': 'unknown', 'type': 'DYNAMIC'}
        assert servers_list._get_preconfigured_password(server) is None


# ---------------------------------------------------------------------------
# Tests – Simulated Zeroconf attack scenario
# ---------------------------------------------------------------------------


class TestZeroconfAttackScenario:
    """End-to-end simulation of the Zeroconf credential exfiltration attack."""

    def test_attacker_advertised_server_no_credential_leak(self, servers_list):
        """Simulate: attacker advertises a fake Zeroconf service with a name matching
        a real server. The browser marks it PROTECTED. Verify no credential is sent."""
        # Simulate a dynamic server entry as Zeroconf would create it
        attacker_server = {
            'key': 'evil-service:61209._glances._tcp.local.',
            'name': 'localhost',  # attacker uses name of a real server
            'ip': '198.51.100.66',  # attacker's IP
            'port': 61209,
            'protocol': 'rpc',
            'username': 'glances',
            'password': '',
            'status': 'UNKNOWN',
            'type': 'DYNAMIC',
        }

        # First probe: no password → URI without credentials
        uri1 = servers_list.get_uri(attacker_server)
        assert '@' not in uri1
        assert '198.51.100.66' in uri1

        # Simulate server responding 401 → status becomes PROTECTED
        attacker_server['password'] = None
        attacker_server['status'] = 'PROTECTED'

        # Second probe: PROTECTED + DYNAMIC → must NOT inject saved password
        uri2 = servers_list.get_uri(attacker_server)
        # password stays None (converted to string in format)
        assert attacker_server['password'] is None
        # No credential hash in URI
        localpwd_hash = servers_list.password.get_hash('localpwd')
        default_hash = servers_list.password.get_hash('defaultpassword')
        assert localpwd_hash not in uri2
        assert default_hash not in uri2



import os
import sys
import time

import psutil

sys.path.insert(0, '../glances')

from glances.globals import disable, enable
from glances.main import GlancesMain
from glances.stats import GlancesStats

core = GlancesMain()
stats = GlancesStats(config=core.get_config(), args=core.get_args())

refresh = 2
iteration = 6
pid = os.getpid()
process = psutil.Process(pid)

print("Check Glances process* plugins CPU consumption overhead")
print("=======================================================")
print()

print(f"Init Glances to make the stats more relevant (wait {iteration * refresh} seconds)")
for i in range(0, iteration):
    stats.update()
    time.sleep(refresh)

print("CPU consumption with process* plugins disable")
for p in ['processcount', 'processlist', 'programlist']:
    disable(stats.get_plugin(p).args, p)
    stats.get_plugin(p).reset()
cpu_start = sum(process.cpu_times()[:2])
for i in range(0, iteration):
    stats.update()
    print(f'{i + 1}/{iteration}')
    time.sleep(refresh)
cpu_end = sum(process.cpu_times()[:2])
cpu_without_sensors = cpu_end - cpu_start
print(f'Glances process consumption (user + kernel) with process* plugins disable: {cpu_without_sensors}')

print("CPU consumption with process* plugins enable")
for p in ['processcount', 'processlist', 'programlist']:
    enable(stats.get_plugin(p).args, p)
    stats.get_plugin(p).reset()
cpu_start = sum(process.cpu_times()[:2])
for i in range(0, iteration):
    stats.update()
    print(f'{i + 1}/{iteration}: {len(stats.get_plugin("processlist").get_raw())} processes')
    time.sleep(refresh)
cpu_end = sum(process.cpu_times()[:2])
cpu_with_sensors = cpu_end - cpu_start
print(f'Glances process consumption (user + kernel) with process* plugins enable: {cpu_with_sensors}')

print(
    f'Percentage of CPU consumption increase with process* plugins enable: \
 {((cpu_with_sensors - cpu_without_sensors) / cpu_without_sensors) * 100:.2f}%'
)



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2022 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Cassandra/Scylla interface class."""

import re
import sys
from datetime import datetime
from numbers import Number

from cassandra import InvalidRequest
from cassandra.auth import PlainTextAuthProvider
from cassandra.cluster import Cluster
from cassandra.util import uuid_from_time

from glances.exports.export import GlancesExport
from glances.logger import logger

_CQL_IDENTIFIER_RE = re.compile(r'^[a-zA-Z][a-zA-Z0-9_]*$')


def _validate_cql_identifier(value, name):
    """Raise ValueError if value is not a safe CQL identifier."""
    if not _CQL_IDENTIFIER_RE.match(str(value)):
        raise ValueError(
            f"Invalid CQL identifier for '{name}': {value!r}. "
            "Only letters, digits, and underscores are allowed, and it must start with a letter."
        )
    return str(value)


class Export(GlancesExport):
    """This class manages the Cassandra/Scylla export module."""

    def __init__(self, config=None, args=None):
        """Init the Cassandra export IF."""
        super().__init__(config=config, args=args)

        # Mandatory configuration keys (additional to host and port)
        self.keyspace = None

        # Optional configuration keys
        self.protocol_version = 3
        self.replication_factor = 2
        self.table = None
        self.username = None
        self.password = None

        # Load the Cassandra configuration file section
        self.export_enable = self.load_conf(
            'cassandra',
            mandatories=['host', 'port', 'keyspace'],
            options=['protocol_version', 'replication_factor', 'table', 'username', 'password'],
        )
        if not self.export_enable:
            sys.exit(2)

        # Validate CQL identifiers to prevent injection via config values
        try:
            self.keyspace = _validate_cql_identifier(self.keyspace, 'keyspace')
            self.table = _validate_cql_identifier(self.table, 'table')
            self.replication_factor = int(self.replication_factor)
            if self.replication_factor < 1:
                raise ValueError("replication_factor must be a positive integer")
        except ValueError as e:
            logger.error(f"Cassandra configuration error: {e}")
            self.export_enable = False
            return

        # Init the Cassandra client
        self.cluster, self.session = self.init()

    def init(self):
        """Init the connection to the Cassandra server."""
        if not self.export_enable:
            return None

        # if username and/or password are not set the connection will try to connect with no auth
        auth_provider = PlainTextAuthProvider(username=self.username, password=self.password)

        # Cluster
        try:
            cluster = Cluster(
                [self.host],
                port=int(self.port),
                protocol_version=int(self.protocol_version),
                auth_provider=auth_provider,
            )
            session = cluster.connect()
        except Exception as e:
            logger.critical(f"Cannot connect to Cassandra cluster '{self.host}:{self.port}' ({e})")
            sys.exit(2)

        # Keyspace
        try:
            session.set_keyspace(self.keyspace)
        except InvalidRequest:
            logger.info(f"Create keyspace {self.keyspace} on the Cassandra cluster")
            c = (
                f"CREATE KEYSPACE {self.keyspace} WITH "
                f"replication = {{ 'class': 'SimpleStrategy', 'replication_factor': '{self.replication_factor}' }}"
            )
            session.execute(c)
            session.set_keyspace(self.keyspace)

        logger.info(
            f"Stats will be exported to Cassandra cluster {cluster.metadata.cluster_name} "
            f"({cluster.metadata.all_hosts()}) in keyspace {self.keyspace}"
        )

        # Table
        try:
            session.execute(
                f"CREATE TABLE {self.table} "
                f"(plugin text, time timeuuid, stat map<text,float>, PRIMARY KEY (plugin, time)) "
                f"WITH CLUSTERING ORDER BY (time DESC)"
            )
        except Exception:
            logger.debug(f"Cassandra table {self.table} already exist")

        return cluster, session

    def export(self, name, columns, points):
        """Write the points to the Cassandra cluster."""
        logger.debug(f"Export {name} stats to Cassandra")

        # Remove non number stats and convert all to float (for Boolean)
        data = {k: float(v) for k, v in zip(columns, points) if isinstance(v, Number)}

        # Write input to the Cassandra table
        try:
            stmt = f"INSERT INTO {self.table} (plugin, time, stat) VALUES (?, ?, ?)"
            query = self.session.prepare(stmt)
            self.session.execute(query, (name, uuid_from_time(datetime.now()), data))
        except Exception as e:
            logger.error(f"Cannot export {name} stats to Cassandra ({e})")

    def exit(self):
        """Close the Cassandra export module."""
        # To ensure all connections are properly closed
        self.session.shutdown()
        self.cluster.shutdown()
        # Call the father method
        super().exit()



#!/usr/bin/env python
#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2025 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Unit tests for the stdout CSV output (--stdout-csv).

Regression tests for issue #3606: when a list-type plugin (e.g. network) gains or
loses items - or an item reappears with a reduced field set (rate fields are not
computed on an interface's first sample) - the data rows must stay aligned with the
header built on the first refresh.
"""

from glances.outputs.glances_stdout_csv import GlancesStdoutCsv


def _make_csv():
    """Return a GlancesStdoutCsv instance without going through __init__."""
    csv = GlancesStdoutCsv.__new__(GlancesStdoutCsv)
    csv.separator = ','
    csv.na = 'N/A'
    csv.header = True
    csv.list_keys = {}
    csv.header_field_names = {}
    return csv


def _iface_full(name, sent, recv):
    """An interface with the full field set (rate/gauge fields present)."""
    return {
        'interface_name': name,
        'key': 'interface_name',
        'bytes_sent': sent,
        'bytes_recv': recv,
        'speed': 1000,
        'alias': None,
        'bytes_all': sent + recv,
        'time_since_update': 2.0,
        'bytes_recv_gauge': recv * 10,
        'bytes_recv_rate_per_sec': float(recv),
        'bytes_sent_gauge': sent * 10,
        'bytes_sent_rate_per_sec': float(sent),
        'bytes_all_gauge': (sent + recv) * 10,
        'bytes_all_rate_per_sec': float(sent + recv),
    }


def _iface_partial(name, sent, recv):
    """An interface on its first sample: rate/gauge fields not computed yet."""
    return {
        'interface_name': name,
        'key': 'interface_name',
        'bytes_sent': sent,
        'bytes_recv': recv,
        'speed': 1000,
        'alias': None,
        'bytes_all': sent + recv,
    }


def _ncols(line):
    stripped = line.rstrip(',')
    return len(stripped.split(',')) if stripped else 0


def test_list_plugin_steady_state():
    """Rows match the header when the interface set is unchanged."""
    csv = _make_csv()
    stat = [_iface_full('eth0', 1, 2), _iface_full('wlan0', 3, 4)]
    header = csv.build_header('network', None, stat)
    data = csv.build_data('network', None, stat)
    assert _ncols(data) == _ncols(header)


def test_list_plugin_interface_removed():
    """A removed interface is N/A-filled so the row stays aligned."""
    csv = _make_csv()
    start = [_iface_full('eth0', 1, 2), _iface_full('wlan0', 3, 4)]
    header = csv.build_header('network', None, start)
    data = csv.build_data('network', None, [_iface_full('eth0', 5, 6)])
    assert _ncols(data) == _ncols(header)
    assert 'N/A' in data


def test_list_plugin_interface_reappears_with_partial_fields():
    """An interface back on its first sample (missing rate fields) is N/A-padded per field."""
    csv = _make_csv()
    start = [_iface_full('eth0', 1, 2), _iface_full('wlan0', 3, 4)]
    header = csv.build_header('network', None, start)
    # wlan0 returns with only the non-rate fields present
    data = csv.build_data('network', None, [_iface_full('eth0', 5, 6), _iface_partial('wlan0', 7, 8)])
    assert _ncols(data) == _ncols(header)
    # the 7 missing wlan0 fields become N/A
    assert data.count('N/A') == 7


def test_list_plugin_interface_added_is_omitted():
    """An interface appearing after export start has no column and is omitted."""
    csv = _make_csv()
    start = [_iface_full('eth0', 1, 2), _iface_full('wlan0', 3, 4)]
    header = csv.build_header('network', None, start)
    data = csv.build_data(
        'network',
        None,
        [_iface_full('eth0', 7, 8), _iface_full('ppp0', 9, 9), _iface_full('wlan0', 10, 11)],
    )
    assert _ncols(data) == _ncols(header)
    assert 'ppp0' not in data


def test_list_plugin_added_keeps_existing_aligned():
    """When a new interface shifts the live order, existing ones stay under their columns."""
    csv = _make_csv()
    start = [_iface_full('eth0', 1, 2), _iface_full('wlan0', 3, 4)]
    csv.build_header('network', None, start)
    data = csv.build_data(
        'network',
        None,
        [_iface_full('eth0', 7, 8), _iface_full('ppp0', 9, 9), _iface_full('wlan0', 10, 11)],
    )
    cells = data.rstrip(',').split(',')
    # wlan0's current bytes_sent (10) must appear under wlan0's block, not eth0's
    assert '10' in cells


def test_dict_plugin_unaffected():
    """Non-list plugins (dict) are unchanged by the fix."""
    csv = _make_csv()
    cpu = {'user': 1.2, 'system': 0.8, 'idle': 98.0}
    header = csv.build_header('cpu', None, cpu)
    data = csv.build_data('cpu', None, cpu)
    assert _ncols(data) == _ncols(header) == 3


def test_attribute_selector_unaffected():
    """A plugin.attribute selector still yields a single aligned column."""
    csv = _make_csv()
    cpu = {'user': 1.2, 'system': 0.8, 'idle': 98.0}
    header = csv.build_header('cpu', 'user', cpu)
    data = csv.build_data('cpu', 'user', cpu)
    assert _ncols(data) == _ncols(header) == 1



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2024 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

from typing import Any, Protocol


class VmsExtension(Protocol):
    def stop(self) -> None:
        raise NotImplementedError

    def update(self, all_tag) -> tuple[dict, list[dict[str, Any]]]:
        raise NotImplementedError



#!/usr/bin/env python
#
# Glances - An eye on your system
#
# SPDX-FileCopyrightText: 2024 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Glances unitary tests suite for the WebUI.

This test uses Selenium to test the Glances WebUI.
Under the wood, it uses the ChromeDriver.

Check your Chrome version with:
    /usr/bin/google-chrome --version
Check your ChromeDriver version with:
    /usr/bin/chromedriver --version

The (major) version should match.

If not, download and install the correct version of ChromeDriver.
https://googlechromelabs.github.io/chrome-for-testing/#stable
"""

import os
import tempfile
import time

import pytest
from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver.common.by import By

SCREENSHOT_RESOLUTIONS = [
    # PC
    (640, 480),
    (800, 600),
    (1024, 768),
    (1600, 900),
    (1280, 1024),
    (1600, 1200),
    (1920, 1200),
    # IPHONE
    (750, 1334),  # 8
    (1080, 1920),  # 8 Plus
    (1242, 2208),  # XS
    (1125, 2436),  # 11 Pro
    (1179, 2556),  # 15
    (1320, 2868),  # 16 Pro Max
    # PIXEL Phone
    (1080, 2400),  # Pixel 7
]


@pytest.fixture(scope="module")
def glances_homepage(web_browser):
    time.sleep(3)
    web_browser.get("http://localhost:61234")
    return web_browser


def test_screenshot(glances_webserver, glances_homepage):
    """
    Test Glances home page screenshot.
    """
    if glances_webserver is None:
        raise AssertionError("Glances webserver is not running")
    for resolution in SCREENSHOT_RESOLUTIONS:
        glances_homepage.set_window_size(*resolution)
        glances_homepage.save_screenshot(
            os.path.join(tempfile.gettempdir(), f"glances-{'-'.join(map(str, list(resolution)))}.png")
        )


def test_loading_time(glances_webserver, glances_homepage):
    """
    Test Glances home page loading time.
    """
    if glances_webserver is None:
        raise AssertionError("Glances webserver is not running")
    navigation_start = glances_homepage.execute_script("return window.performance.timing.navigationStart")
    response_start = glances_homepage.execute_script("return window.performance.timing.responseStart")
    dom_complete = glances_homepage.execute_script("return window.performance.timing.domComplete")
    backend_perf = response_start - navigation_start
    frontend_perf = dom_complete - response_start
    if backend_perf >= 2000:
        raise AssertionError(f"Backend performance is too slow: {backend_perf}ms (limit is 2000ms)")
    if frontend_perf >= 2000:
        raise AssertionError(f"Frontend performance is too slow: {frontend_perf}ms (limit is 2000ms)")


def test_title(glances_webserver, glances_homepage):
    """
    Test Glances home page title.
    """
    if glances_webserver is None:
        raise AssertionError("Glances webserver is not running")
    if "Glances" not in glances_homepage.title:
        raise AssertionError(f"Expected 'Glances' in title, but got '{glances_homepage.title}'")


def test_plugins(glances_webserver, glances_homepage):
    """
    Test Glances defaults plugins.
    """
    if glances_webserver is None:
        raise AssertionError("Glances webserver is not running")
    for plugin in [
        "system",
        "now",
        "uptime",
        "quicklook",
        "cpu",
        "mem",
        "memswap",
        "load",
        "network",
        "diskio",
        "fs",
        "sensors",
        "processcount",
        "processlist",
    ]:
        if plugin == 'sensors':
            try:
                assert glances_homepage.find_element(By.ID, plugin) is not None
            except NoSuchElementException:
                # Sensors can be hidden on VM
                pass
        else:
            assert glances_homepage.find_element(By.ID, plugin) is not None



import json

from glances.outputs.glances_curses import _GlancesCurses

print(
    json.dumps(
        {
            "topMenu": list(_GlancesCurses._top),
            "leftMenu": [p for p in _GlancesCurses._left_sidebar if p != "now"],
        },
        indent=4,
    )
)



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2022 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only

"""Podman Extension unit for Glances' Containers plugin."""

import time
from datetime import datetime
from typing import Any

import psutil

from glances.globals import nativestr, pretty_date, replace_special_chars, string_value_to_float
from glances.logger import logger
from glances.stats_streamer import ThreadedIterableStreamer

# Podman library (optional and Linux-only)
# https://pypi.org/project/podman/
try:
    from podman import PodmanClient
except Exception as e:
    disable_plugin_podman = True
    # Display debug message if import KeyError
    logger.warning(f"Error loading Podman deps Lib. Podman feature in the Containers plugin is disabled ({e})")
else:
    disable_plugin_podman = False


class PodmanContainerStatsFetcher:
    MANDATORY_FIELDS = ["CPU", "MemUsage", "MemLimit", "BlockInput", "BlockOutput"]

    def __init__(self, container):
        self._container = container

        # Full inspection to get limits (see issue #1152: List API doesn't include HostConfig)
        try:
            self._inspect_data = self._container.inspect()
        except Exception:
            self._inspect_data = {}

        # Previous stats are stored in the self._old_computed_stats variable
        # We store time data to enable rate calculations to avoid complexity for consumers of the APIs exposed.
        self._old_computed_stats = {}

        # Last time when output stats (results) were computed
        self._last_stats_computed_time = 0

        # Threaded Streamer
        stats_iterable = container.stats(decode=True)
        self._streamer = ThreadedIterableStreamer(stats_iterable, initial_stream_value={})

    def stop(self):
        self._streamer.stop()

    def get_streamed_stats(self) -> dict[str, Any]:
        stats = self._streamer.stats
        if stats is None or stats.get("Error", False):
            logger.error(f"containers (Podman) Container({self._container.id}): Stats fetching failed")
            logger.debug(f"containers (Podman) Container({self._container.id}): ", stats)
            return None

        return stats["Stats"][0]

    @property
    def activity_stats(self) -> dict[str, Any]:
        """Activity Stats

        Each successive access of activity_stats will cause computation of activity_stats
        """
        computed_activity_stats = self._compute_activity_stats()
        self._old_computed_stats = computed_activity_stats
        self._last_stats_computed_time = time.time()
        return computed_activity_stats

    def _compute_activity_stats(self) -> dict[str, dict[str, Any]]:
        stats = {"cpu": {}, "memory": {}, "io": {}, "network": {}}
        api_stats = self.get_streamed_stats()

        # Glances breaks if Podman container is started while it is running See #3199
        if api_stats is None:
            # If stats fetching failed, return empty stats
            logger.error(f"containers (Podman) Container({self._container.id}): Failed to fetch stats")
            return stats

        if any(field not in api_stats for field in self.MANDATORY_FIELDS) or (
            "Network" not in api_stats and any(k not in api_stats for k in ["NetInput", "NetOutput"])
        ):
            logger.error(f"containers (Podman) Container({self._container.id}): Missing mandatory fields")
            return stats

        try:
            stats["cpu"]["total"] = api_stats["CPU"]
            # CPU limit
            host_config = self._inspect_data.get('HostConfig')
            nano_cpus = host_config.get('NanoCpus', 0)
            if nano_cpus > 0:
                stats['cpu']['limit'] = nano_cpus / 1e9
            else:
                cpu_quota = host_config.get('CpuQuota', 0)
                cpu_period = host_config.get('CpuPeriod', 0)
                if cpu_quota > 0 and cpu_period > 0:
                    stats['cpu']['limit'] = cpu_quota / cpu_period
                else:
                    # Fallback to the number of available cores
                    stats['cpu']['limit'] = float(psutil.cpu_count())

            stats["memory"]["usage"] = api_stats["MemUsage"]
            stats["memory"]["limit"] = api_stats["MemLimit"]

            stats["io"]["ior"] = api_stats["BlockInput"]
            stats["io"]["iow"] = api_stats["BlockOutput"]
            stats["io"]["time_since_update"] = 1
            # Hardcode `time_since_update` to 1 as podman already sends at the same fixed rate per second

            if "Network" not in api_stats:
                # For podman rooted mode
                stats["network"]["rx"] = api_stats["NetInput"]
                stats["network"]["tx"] = api_stats["NetOutput"]
                stats["network"]["time_since_update"] = 1
                # Hardcode to 1 as podman already sends at the same fixed rate per second
            elif api_stats["Network"] is not None:
                # api_stats["Network"] can be None if the infra container of the pod is killed
                # For podman in rootless mode
                stats["network"] = {
                    "cumulative_rx": sum(interface["RxBytes"] for interface in api_stats["Network"].values()),
                    "cumulative_tx": sum(interface["TxBytes"] for interface in api_stats["Network"].values()),
                }
                # Using previous stats to calculate rates
                old_network_stats = self._old_computed_stats.get("network")
                if old_network_stats:
                    stats["network"]["time_since_update"] = round(self.time_since_update)
                    stats["network"]["rx"] = stats["network"]["cumulative_rx"] - old_network_stats["cumulative_rx"]
                    stats["network"]["tx"] = stats["network"]["cumulative_tx"] - old_network_stats["cumulative_tx"]

        except ValueError as e:
            logger.error(
                f"containers (Podman) Container({self._container.id}): Non float stats values found",
                e,
            )

        return stats

    @property
    def time_since_update(self) -> float:
        # In case no update (at startup), default to 1
        return max(1, self._streamer.last_update_time - self._last_stats_computed_time)


class PodmanPodStatsFetcher:
    def __init__(self, pod_manager):
        self._pod_manager = pod_manager

        # Threaded Streamer
        # Temporary patch to get podman extension working
        stats_iterable = (pod_manager.stats(decode=True) for _ in iter(int, 1))
        # WARNING: Podman API doesn't specify the rate at which stats are sent, so we set it to 1 second
        # to avoid overloading the system with stats calculations. With a lot of pods, this can cause some
        # delay in stats display but it is better than overloading the system.
        self._streamer = ThreadedIterableStreamer(stats_iterable, initial_stream_value={}, sleep_duration=1)

    def _log_debug(self, msg, exception=None):
        logger.debug(f"containers (Podman): Pod Manager - {msg} ({exception})")
        logger.debug(self._streamer.stats)

    def stop(self):
        self._streamer.stop()

    @property
    def activity_stats(self):
        result_stats = {}
        container_stats = self._streamer.stats
        for stat in container_stats:
            io_stats = self._get_io_stats(stat)
            cpu_stats = self._get_cpu_stats(stat)
            memory_stats = self._get_memory_stats(stat)
            network_stats = self._get_network_stats(stat)

            computed_stats = {
                "name": stat["Name"],
                "cid": stat["CID"],
                "pod_id": stat["Pod"],
                "io": io_stats or {},
                "memory": memory_stats or {},
                "network": network_stats or {},
                "cpu": cpu_stats or {},
            }
            result_stats[stat["CID"]] = computed_stats

        return result_stats

    def _get_cpu_stats(self, stats: dict) -> dict | None:
        """Return the container CPU usage.

        Output: a dict {'total': 1.49}
        """
        if "CPU" not in stats:
            self._log_debug("Missing CPU usage fields")
            return None

        cpu_usage = string_value_to_float(stats["CPU"].rstrip("%"))
        return {"total": cpu_usage}

    def _get_memory_stats(self, stats) -> dict | None:
        """Return the container MEMORY.

        Output: a dict {'usage': ..., 'limit': ...}
        """
        if "MemUsage" not in stats or "/" not in stats["MemUsage"]:
            self._log_debug("Missing MEM usage fields")
            return None

        memory_usage_str = stats["MemUsage"]
        usage_str, limit_str = memory_usage_str.split("/")

        try:
            usage = string_value_to_float(usage_str)
            limit = string_value_to_float(limit_str)
        except ValueError as e:
            self._log_debug("Compute MEM usage failed", e)
            return None

        return {"usage": usage, "limit": limit, "inactive_file": 0}

    def _get_network_stats(self, stats) -> dict | None:
        """Return the container network usage using the Docker API (v1.0 or higher).

        Output: a dict {'time_since_update': 3000, 'rx': 10, 'tx': 65}.
        with:
            time_since_update: number of seconds elapsed between the latest grab
            rx: Number of bytes received
            tx: Number of bytes transmitted
        """
        if "NetIO" not in stats or "/" not in stats["NetIO"]:
            self._log_debug("Compute MEM usage failed")
            return None

        net_io_str = stats["NetIO"]
        rx_str, tx_str = net_io_str.split("/")

        try:
            rx = string_value_to_float(rx_str)
            tx = string_value_to_float(tx_str)
        except ValueError as e:
            self._log_debug("Compute MEM usage failed", e)
            return None

        # Hardcode `time_since_update` to 1 as podman docs don't specify the rate calculation procedure
        return {"rx": rx, "tx": tx, "time_since_update": 1}

    def _get_io_stats(self, stats) -> dict | None:
        """Return the container IO usage using the Docker API (v1.0 or higher).

        Output: a dict {'time_since_update': 3000, 'ior': 10, 'iow': 65}.
        with:
            time_since_update: number of seconds elapsed between the latest grab
            ior: Number of bytes read
            iow: Number of bytes written
        """
        if "BlockIO" not in stats or "/" not in stats["BlockIO"]:
            self._log_debug("Missing BlockIO usage fields")
            return None

        block_io_str = stats["BlockIO"]
        ior_str, iow_str = block_io_str.split("/")

        try:
            ior = string_value_to_float(ior_str)
            iow = string_value_to_float(iow_str)
        except ValueError as e:
            self._log_debug("Compute BlockIO usage failed", e)
            return None

        # Hardcode `time_since_update` to 1 as podman docs don't specify the rate calculation procedure
        return {"ior": ior, "iow": iow, "time_since_update": 1}


class PodmanExtension:
    """Glances' Containers Plugin's Docker Extension unit"""

    CONTAINER_ACTIVE_STATUS = ["running", "healthy", "paused"]

    def __init__(self, podman_sock):
        self.disable = disable_plugin_podman
        if self.disable:
            raise Exception("Missing libs required to run Podman Extension (Containers)")

        self.display_error = True

        self.client = None
        self.ext_name = "containers (Podman)"
        self.podman_sock = podman_sock
        self.pods_stats_fetcher = None
        self.container_stats_fetchers = {}

        # Issue #3559: cache the (immutable) image tags per container id to avoid
        # one inspect_image API call per container on every refresh.
        self.image_cache = {}

        self.connect()

    def connect(self):
        """Connect to Podman."""
        try:
            self.client = PodmanClient(base_url=self.podman_sock)
            # PodmanClient works lazily, so make a ping to determine if socket is open
            self.client.ping()
        except Exception as e:
            logger.debug(f"{self.ext_name} plugin - Can't connect to Podman ({e})")
            self.client = None
            self.disable = True

    def update_version(self):
        # Long and not useful anymore because the information is no more displayed in UIs
        # return self.client.version()
        return {}

    def stop(self) -> None:
        # Stop all streaming threads
        for t in self.container_stats_fetchers.values():
            t.stop()

        if self.pods_stats_fetcher:
            self.pods_stats_fetcher.stop()

    def update(self, all_tag) -> tuple[dict, list[dict[str, Any]]]:
        """Update Podman stats using the input method."""

        if not self.client or self.disable:
            return {}, []

        version_stats = self.update_version()

        # Update current containers list
        try:
            # Issue #1152: Podman module doesn't export details about stopped containers
            # The Containers/all key of the configuration file should be set to True
            containers = self.client.containers.list(all=all_tag)
            if not self.pods_stats_fetcher:
                self.pods_stats_fetcher = PodmanPodStatsFetcher(self.client.pods)
            self.display_error = True
        except Exception as e:
            if self.display_error:
                logger.error(f"{self.ext_name} plugin - Can't get containers list ({e})")
                self.display_error = False
            else:
                logger.debug(f"{self.ext_name} plugin - Can't get containers list ({e})")
            return version_stats, []

        # Start new thread for new container
        for container in containers:
            if container.id not in self.container_stats_fetchers:
                # StatsFetcher did not exist in the internal dict
                # Create it, add it to the internal dict
                logger.debug(f"{self.ext_name} plugin - Create thread for container {container.id[:12]}")
                self.container_stats_fetchers[container.id] = PodmanContainerStatsFetcher(container)

        # Stop threads for non-existing containers
        current_ids = {c.id for c in containers}
        absent_containers = set(self.container_stats_fetchers.keys()) - current_ids
        for container_id in absent_containers:
            # Stop the StatsFetcher
            logger.debug(f"{self.ext_name} plugin - Stop thread for old container {container_id[:12]}")
            self.container_stats_fetchers[container_id].stop()
            # Delete the StatsFetcher from the dict
            del self.container_stats_fetchers[container_id]

        # Drop image cache entries for non-existing containers (issue #3559)
        for container_id in set(self.image_cache) - current_ids:
            del self.image_cache[container_id]

        # Get stats for all containers
        container_stats = [self.generate_stats(container) for container in containers]

        pod_stats = self.pods_stats_fetcher.activity_stats
        for stats in container_stats:
            if stats["id"][:12] in pod_stats:
                stats["pod_name"] = pod_stats[stats["id"][:12]]["name"]
                stats["pod_id"] = pod_stats[stats["id"][:12]]["pod_id"]

        return version_stats, container_stats

    @property
    def key(self) -> str:
        """Return the key of the list."""
        return "name"

    def _get_image(self, container):
        """Return the container image tags, cached per container id (issue #3559).

        `container.image` performs an inspect_image API call. The image of a container
        is immutable for its lifetime, so the result is cached to avoid one synchronous
        API call per container on every refresh.
        """
        if container.id not in self.image_cache:
            # Access container.image only once (it triggers an inspect_image API call)
            tags = container.image.tags
            self.image_cache[container.id] = ",".join(tags if tags else [])
        return self.image_cache[container.id]

    def generate_stats(self, container) -> dict[str, Any]:
        # Init the stats for the current container
        stats = {
            "key": self.key,
            "name": nativestr(container.name),
            "id": container.id,
            "image": self._get_image(container),
            "status": container.attrs["State"],
            "created": container.attrs["Created"],
            "command": container.attrs.get("Command") or [],
            "io": {},
            "cpu": {},
            "memory": {},
            "network": {},
            "io_rx": None,
            "io_wx": None,
            "cpu_percent": None,
            "memory_percent": None,
            "network_rx": None,
            "network_tx": None,
            "ports": "",
            "uptime": None,
        }

        stats_fetcher = self.container_stats_fetchers[container.id]
        activity_stats = stats_fetcher.activity_stats
        stats.update(activity_stats)

        # Additional fields
        stats["cpu_percent"] = stats["cpu"].get("total")
        stats["cpu_limit"] = stats["cpu"].get("limit")
        stats["memory_usage"] = stats["memory"].get("usage")
        if stats["memory"].get("cache") is not None:
            stats["memory_usage"] -= stats["memory"]["cache"]
        stats["memory_inactive_file"] = stats["memory"].get("inactive_file")
        stats["memory_limit"] = stats["memory"].get("limit")

        if all(k in stats["io"] for k in ("ior", "iow", "time_since_update")):
            stats["io_rx"] = stats["io"]["ior"] // stats["io"]["time_since_update"]
            stats["io_wx"] = stats["io"]["iow"] // stats["io"]["time_since_update"]

        if all(k in stats["network"] for k in ("rx", "tx", "time_since_update")):
            stats["network_rx"] = stats["network"]["rx"] // stats["network"]["time_since_update"]
            stats["network_tx"] = stats["network"]["tx"] // stats["network"]["time_since_update"]

        started_at = datetime.fromtimestamp(container.attrs["StartedAt"])
        stats["uptime"] = pretty_date(started_at)

        # Manage special chars in command (see issue#2733)
        stats["command"] = replace_special_chars(" ".join(stats["command"]))

        # Manage ports (see issue#2054)
        if hasattr(container, "ports"):
            stats["ports"] = ",".join(
                [
                    (f'{container.ports[cp][0]["HostPort"]}->{cp}' if container.ports[cp] else f"{cp}")
                    for cp in container.ports
                ]
            )

        return stats


# End of file



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2022 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""RESTful interface class."""

from requests import post

from glances.exports.export import GlancesExport
from glances.globals import listkeys
from glances.logger import logger


class Export(GlancesExport):
    """This class manages the RESTful export module.
    Be aware that stats will be exported in one big POST request"""

    def __init__(self, config=None, args=None):
        """Init the RESTful export IF."""
        super().__init__(config=config, args=args)

        # Mandatory configuration keys (additional to host and port)
        self.protocol = None
        self.path = None

        # Load the RESTful section in the configuration file
        self.export_enable = self.load_conf('restful', mandatories=['host', 'port', 'protocol', 'path'])
        if not self.export_enable:
            exit('Missing RESTFUL config')

        # Init the stats buffer
        # It's a dict of stats
        self.buffer = {}

        # Init the Statsd client
        self.client = self.init()

    def init(self):
        """Init the connection to the RESTful server."""
        if not self.export_enable:
            return None
        # Build the RESTful URL where the stats will be posted
        url = f'{self.protocol}://{self.host}:{self.port}{self.path}'
        logger.info(f"Stats will be exported to the RESTful endpoint {url}")
        return url

    def export(self, name, columns, points):
        """Export the stats to the Statsd server."""
        if name == self.last_exported_list()[0] and self.buffer != {}:
            # One complete loop have been done
            logger.debug(f"Export stats ({listkeys(self.buffer)}) to RESTful endpoint ({self.client})")
            # Export stats
            post(self.client, json=self.buffer, allow_redirects=True, timeout=15)
            # Reset buffer
            self.buffer = {}

        # Add current stat to the buffer
        self.buffer[name] = dict(zip(columns, points))



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2022 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Per-CPU plugin."""

from glances.cpu_percent import cpu_percent
from glances.globals import BSD, LINUX, MACOS, WINDOWS
from glances.plugins.plugin.model import GlancesPluginModel

# Fields description
# description: human readable description
# short_name: shortname to use un UI
# unit: unit type
# rate: is it a rate ? If yes, // by time_since_update when displayed,
# min_symbol: Auto unit should be used if value > than 1 'X' (K, M, G)...
fields_description = {
    'cpu_number': {
        'description': 'CPU number',
    },
    'total': {
        'description': 'Sum of CPU percentages (except idle) for current CPU number.',
        'unit': 'percent',
    },
    'system': {
        'description': 'Percent time spent in kernel space. System CPU time is the \
time spent running code in the Operating System kernel.',
        'unit': 'percent',
    },
    'user': {
        'description': 'CPU percent time spent in user space. \
User CPU time is the time spent on the processor running your program\'s code (or code in libraries).',
        'unit': 'percent',
    },
    'iowait': {
        'description': '*(Linux)*: percent time spent by the CPU waiting for I/O \
operations to complete.',
        'unit': 'percent',
    },
    'idle': {
        'description': 'percent of CPU used by any program. Every program or task \
that runs on a computer system occupies a certain amount of processing \
time on the CPU. If the CPU has completed all tasks it is idle.',
        'unit': 'percent',
    },
    'irq': {
        'description': '*(Linux and BSD)*: percent time spent servicing/handling \
hardware/software interrupts. Time servicing interrupts (hardware + \
software).',
        'unit': 'percent',
    },
    'nice': {
        'description': '*(Unix)*: percent time occupied by user level processes with \
a positive nice value. The time the CPU has spent running users\' \
processes that have been *niced*.',
        'unit': 'percent',
    },
    'steal': {
        'description': '*(Linux)*: percentage of time a virtual CPU waits for a real \
CPU while the hypervisor is servicing another virtual processor.',
        'unit': 'percent',
    },
    'guest': {
        'description': '*(Linux)*: percent of time spent running a virtual CPU for \
guest operating systems under the control of the Linux kernel.',
        'unit': 'percent',
    },
    'guest_nice': {
        'description': '*(Linux)*: percent of time spent running a niced guest (virtual CPU).',
        'unit': 'percent',
    },
    'softirq': {
        'description': '*(Linux)*: percent of time spent handling software interrupts.',
        'unit': 'percent',
    },
    'dpc': {
        'description': '*(Windows)*: percent of time spent handling deferred procedure calls.',
        'unit': 'percent',
    },
    'interrupt': {
        'description': '*(Windows)*: percent of time spent handling software interrupts.',
        'unit': 'percent',
    },
}

# Define the history items list
items_history_list = [
    {'name': 'user', 'description': 'User CPU usage', 'y_unit': '%'},
    {'name': 'system', 'description': 'System CPU usage', 'y_unit': '%'},
]


class PercpuPlugin(GlancesPluginModel):
    """Glances per-CPU plugin.

    'stats' is a list of dictionaries that contain the utilization percentages
    for each CPU.
    """

    def __init__(self, args=None, config=None):
        """Init the plugin."""
        super().__init__(
            args=args,
            config=config,
            items_history_list=items_history_list,
            stats_init_value=[],
            fields_description=fields_description,
        )

        # We want to display the stat in the curse interface
        self.display_curse = True

        # Manage the maximum number of CPU to display (related to enhancement request #2734)
        if config:
            self.max_cpu_display = config.get_int_value('percpu', 'max_cpu_display', 4)
        else:
            self.max_cpu_display = 4

    def get_key(self):
        """Return the key of the list."""
        return 'cpu_number'

    @GlancesPluginModel._check_decorator
    @GlancesPluginModel._log_result_decorator
    def update(self):
        """Update per-CPU stats using the input method."""
        # Grab per-CPU stats using psutil's
        if self.input_method == 'local':
            stats = cpu_percent.get_percpu()
        else:
            # Update stats using SNMP
            stats = self.get_init_value()

        # Update the stats
        self.stats = stats

        return self.stats

    def define_headers_from_os(self):
        base = ['user', 'system']

        if LINUX:
            base += ['iowait', 'idle', 'irq', 'nice', 'steal', 'guest']
        elif MACOS:
            base += ['idle', 'nice']
        elif BSD:
            base += ['idle', 'irq', 'nice']
        elif WINDOWS:
            base += ['dpc', 'interrupt']

        return base

    def maybe_build_string_msg(self, header, return_):
        if self.is_disabled('quicklook'):
            msg = '{:5}'.format('CPU')
            return_.append(self.curse_add_line(msg, "TITLE"))
            header.insert(0, 'total')

        return (header, return_)

    def display_cpu_stats_per_line(self, header, return_):
        for stat in header:
            msg = f'{stat:>7}'
            return_.append(self.curse_add_line(msg))

        return return_

    def manage_max_cpu_to_display(self):
        if len(self.stats) > self.max_cpu_display:
            # sort and display top 'n'
            percpu_list = sorted(self.stats, key=lambda x: x['total'], reverse=True)
        else:
            percpu_list = self.stats

        return percpu_list

    def display_cpu_header_in_columns(self, cpu, return_):
        return_.append(self.curse_new_line())
        if self.is_disabled('quicklook'):
            try:
                cpu_id = cpu[cpu['key']]
                if cpu_id < 10:
                    msg = f'CPU{cpu_id:1} '
                else:
                    msg = f'{cpu_id:4} '
            except TypeError:
                # TypeError: string indices must be integers (issue #1027)
                msg = '{:4} '.format('?')
            return_.append(self.curse_add_line(msg))

        return return_

    def display_cpu_stats_in_columns(self, cpu, header, return_):
        for stat in header:
            try:
                msg = f'{cpu[stat]:6.1f}%'
            except TypeError:
                msg = '{:>6}%'.format('?')
            return_.append(self.curse_add_line(msg, self.get_alert(cpu[stat], header=stat)))

        return return_

    def summarize_all_cpus_not_displayed(self, percpu_list, header, return_):
        if len(self.stats) > self.max_cpu_display:
            return_.append(self.curse_new_line())
            if self.is_disabled('quicklook'):
                return_.append(self.curse_add_line('CPU* '))

            for stat in header:
                percpu_stats = [i[stat] for i in percpu_list[0 : self.max_cpu_display]]
                cpu_stat = sum(percpu_stats) / len(percpu_stats)
                try:
                    msg = f'{cpu_stat:6.1f}%'
                except TypeError:
                    msg = '{:>6}%'.format('?')
                return_.append(self.curse_add_line(msg, self.get_alert(cpu_stat, header=stat)))

        return return_

    def msg_curse(self, args=None, max_width=None):
        """Return the list of dict to display in the curse interface."""

        # Init the return message
        return_ = []

        # Only process if stats exist...
        missing = [not self.stats, not self.args.percpu, self.is_disabled()]
        if any(missing):
            return return_

        # Define the headers based on OS
        header = self.define_headers_from_os()

        # Build the string message
        header, return_ = self.maybe_build_string_msg(header, return_)

        # Per CPU stats displayed per line
        return_ = self.display_cpu_stats_per_line(header, return_)

        # Manage the maximum number of CPU to display (related to enhancement request #2734)
        percpu_list = self.manage_max_cpu_to_display()

        # Per CPU stats displayed per column
        for cpu in percpu_list[0 : self.max_cpu_display]:
            header_added = self.display_cpu_header_in_columns(cpu, return_)
            stats_added = self.display_cpu_stats_in_columns(cpu, header, header_added)

        # Add a new line with sum of all others CPU
        return self.summarize_all_cpus_not_displayed(percpu_list, header, stats_added)



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2024 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Alert plugin."""

from datetime import datetime
from functools import reduce

from glances.events_list import glances_events
from glances.plugins.plugin.model import GlancesPluginModel

# {
#     "begin": "begin",
#     "end": "end",
#     "state": "WARNING|CRITICAL",
#     "type": "CPU|LOAD|MEM",
#     "max": MAX,
#     "avg": AVG,
#     "min": MIN,
#     "sum": SUM,
#     "count": COUNT,
#     "top": [top3 process list],
#     "desc": "Processes description",
#     "sort": "top sort key"
#     "global": "global alert message"
# }
# Fields description
# description: human readable description
# short_name: shortname to use un UI
# unit: unit type
# rate: is it a rate ? If yes, // by time_since_update when displayed,
# min_symbol: Auto unit should be used if value > than 1 'X' (K, M, G)...
fields_description = {
    'begin': {
        'description': 'Begin timestamp of the event',
        'unit': 'timestamp',
    },
    'end': {
        'description': 'End timestamp of the event (or -1 if ongoing)',
        'unit': 'timestamp',
    },
    'state': {
        'description': 'State of the event (WARNING|CRITICAL)',
        'unit': 'string',
    },
    'type': {
        'description': 'Type of the event (CPU|LOAD|MEM)',
        'unit': 'string',
    },
    'max': {
        'description': 'Maximum value during the event period',
        'unit': 'float',
    },
    'avg': {
        'description': 'Average value during the event period',
        'unit': 'float',
    },
    'min': {
        'description': 'Minimum value during the event period',
        'unit': 'float',
    },
    'sum': {
        'description': 'Sum of the values during the event period',
        'unit': 'float',
    },
    'count': {
        'description': 'Number of values during the event period',
        'unit': 'int',
    },
    'top': {
        'description': 'Top 3 processes name during the event period',
        'unit': 'list',
    },
    'desc': {
        'description': 'Description of the event',
        'unit': 'string',
    },
    'sort': {
        'description': 'Sort key of the top processes',
        'unit': 'string',
    },
    'global_msg': {
        'description': 'Global alert message',
        'unit': 'string',
    },
}


class AlertPlugin(GlancesPluginModel):
    """Glances alert plugin.

    Only for display.
    """

    def __init__(self, args=None, config=None):
        """Init the plugin."""
        super().__init__(args=args, config=config, stats_init_value=[], fields_description=fields_description)

        # We want to display the stat in the curse interface
        self.display_curse = True

        # Set the message position
        self.align = 'bottom'

        # Set the maximum number of events to display
        if config is not None and (config.has_section('alert') or config.has_section('alerts')):
            glances_events.set_max_events(config.get_int_value('alert', 'max_events', default=10))
            glances_events.set_min_duration(config.get_int_value('alert', 'min_duration', default=6))
            glances_events.set_min_interval(config.get_int_value('alert', 'min_interval', default=6))
        else:
            glances_events.set_max_events(10)
            glances_events.set_min_duration(6)
            glances_events.set_min_interval(6)

    def update(self):
        """Nothing to do here. Just return the global glances_log."""
        # Set the stats to the glances_events
        self.stats = glances_events.get()

    def build_hdr_msg(self, ret):
        def cond(elem):
            return elem['end'] == -1 and 'global_msg' in elem

        global_message = [elem['global_msg'] for elem in self.stats if cond(elem)]
        title = global_message[0] if global_message else "EVENTS history"

        ret.append(self.curse_add_line(title, "TITLE"))

        return ret

    def add_new_line(self, ret, alert):
        ret.append(self.curse_new_line())

        return ret

    def add_start_time(self, ret, alert):
        timezone = datetime.now().astimezone().tzinfo
        alert_dt = datetime.fromtimestamp(alert['begin'], tz=timezone)
        ret.append(self.curse_add_line(alert_dt.strftime("%Y-%m-%d %H:%M:%S(%z)")))

        return ret

    def add_duration(self, ret, alert):
        if alert['end'] > 0:
            # If finished display duration
            end = datetime.fromtimestamp(alert['end'])
            begin = datetime.fromtimestamp(alert['begin'])
            msg = f' ({end - begin})'
        else:
            msg = ' (ongoing)'
        ret.append(self.curse_add_line(msg))
        ret.append(self.curse_add_line(" - "))

        return ret

    def add_infos(self, ret, alert):
        if alert['end'] > 0:
            # If finished do not display status
            msg = '{} on {}'.format(alert['state'], alert['type'])
            ret.append(self.curse_add_line(msg))
        else:
            msg = str(alert['type'])
            ret.append(self.curse_add_line(msg, decoration=alert['state']))

        return ret

    def add_min_mean_max(self, ret, alert):
        if self.approx_equal(alert['min'], alert['max'], tolerance=0.1):
            msg = ' ({:.1f})'.format(alert['avg'])
        else:
            msg = ' (Min:{:.1f} Mean:{:.1f} Max:{:.1f})'.format(alert['min'], alert['avg'], alert['max'])
        ret.append(self.curse_add_line(msg))

        return ret

    def add_top_proc(self, ret, alert):
        top_process = ', '.join(alert['top'])
        if top_process != '':
            msg = f': {top_process}'
            ret.append(self.curse_add_line(msg))

        return ret

    def loop_over_alert(self, init, alert):
        steps = [
            self.add_new_line,
            self.add_start_time,
            self.add_duration,
            self.add_infos,
            self.add_min_mean_max,
            self.add_top_proc,
        ]

        return reduce(lambda ret, step: step(ret, alert), steps, init)

    def msg_curse(self, args=None, max_width=None):
        """Return the dict to display in the curse interface."""
        # Init the return message
        init = []

        # Only process if display plugin enable...
        if not self.stats or self.is_disabled():
            return init

        return reduce(self.loop_over_alert, self.stats, self.build_hdr_msg(init))

    def approx_equal(self, a, b, tolerance=0.0):
        """Compare a with b using the tolerance (if numerical)."""
        if str(int(a)).isdigit() and str(int(b)).isdigit():
            return abs(a - b) <= max(abs(a), abs(b)) * tolerance
        return a == b



import sys
import time

sys.path.insert(0, '../glances')

###########

# from glances.cpu_percent import cpu_percent

# for _ in range(0, 5):
#     print([i['total'] for i in cpu_percent.get_percpu()])
#     time.sleep(2)

###########

from glances.main import GlancesMain
from glances.stats import GlancesStats

core = GlancesMain()
stats = GlancesStats(config=core.get_config(), args=core.get_args())

for _ in range(0, 5):
    stats.update()
    print([i['total'] for i in stats.get_plugin('percpu').get_raw()])
    time.sleep(2)
